# frozen_string_literal: true

require 'digest'

# manage your asset names
#
# a config directory contains a file for each managed asset using its generic name. In the file
# is the unique stamp that is used to find the name of the latest version of the asset.
# config format consists of a filestamp part and md5 hash part
#
# ensure that your config dir exists (default config/assets)
# ensure that you destination dir exists ( eg public/assets )
#
#  then use something like this to compile your assets...
#
# ASSETS = ['main.js', 'main.css']
# ASSETS.each do |name|
#
#   str = environment[name].to_s
#
#   if File.extname(name) == ".css"
#     engine = Sass::Engine.new(str, :syntax => :scss, :style => :compressed)
#     str = engine.render
#   end
#
#   AssetManager.if_modified(name,str,:rewrite=>true) do |a|
#
#     filename = File.join(ROOT,"www","assets",a.newname)
#     puts "writing #{name} to #{filename}"
#     File.write filename,str
#
#     File.unlink File.join(ROOT,"www","assets",a.oldname) if a.oldname
#   end
#
# end
#
#
# then within your erb template or whatever ...
#
# <script src="<%= AssetManager.asset_path('/assets/main.js',flag) %>" type="text/javascript"></script>
#
# where flag indicates if the static compiled version should be used ( ie with hash) or the plain name ( served by sprockets pipeline)

module Blix
  class AssetManager

    AssetInfo = Struct.new(:newname, :oldname)
    class << self

      def config_dir
        @config_dir || 'config/assets'
      end

      attr_writer :config_dir

      # the name of the file to write to
      def filename(name)
        name
      end

      def asset_root
        @asset_root || 'assets'
      end

      attr_writer :asset_root

      # write the compiled asset here
      def dest_path(name)
        File.join(asset_dir, filename(name))
      end

      # write the config info here
      def config_path(name)
        File.join(config_dir, filename(name) + '.conf')
      end

      # yield the old and new name to a block if
      # the asset has been modified
      def if_modified(destination, name, data, opts = {})
        new_hash = Digest::MD5.hexdigest data

        ext = File.extname(name)[1..-1]
        base = name.split('.')[0]
        confname = base + '.' + ext
        config = get_config(confname)

        if !config ||
           (config[:hash] != new_hash || (config[:destination] != destination)) ||
           opts[:version] && (opts[:version] != config[:version])

          stamp = get_filestamp
          info = AssetInfo.new
          info.newname = "#{base}-#{opts[:version] || stamp}.#{ext}"
          info.oldname = config && "#{base}-#{config[:version] || config[:stamp]}.#{ext}"
          set_config(destination, confname, new_hash, stamp, opts[:version])
          yield(info)

        elsif opts[:rewrite] # rewrite the same data to the same asset as before

          info = AssetInfo.new
          info.newname = config && "#{base}-#{config[:version] || config[:stamp]}.#{ext}"
          info.oldname = nil
          yield(info)

        end
      end

      # generate a unique suffix for the file
      def get_filestamp
        now = Time.now
        str = '%X' % now.to_i
        str += '%X' % now.usec
        str += '%X' % rand(9999)
        str
      end

      # get config data from file or nil if file does not exist
      def get_config(name)
        return nil unless File.exist? config_path(name)

        data = File.read config_path(name)
        parts = data.split('|')
        out = { :hash => parts[1], :stamp => parts[0], :destination => parts[3] }
        out[:version] = parts[2] unless parts[2].empty?
        out
      end

      # set the config data
      def set_config(destination, name, hash, stamp, version = nil)
        File.write(config_path(name), stamp + '|' + hash + '|' + version.to_s + '|' + destination)
        true
      end

      # cache asset name store
      def cache
        @cache ||= {}
      end

      # retrieve and cache the full asset name
      def get_asset_name(name)
        cache[name] ||= begin
          config = get_config(name)
          raise "ERROR : config file for asset:#{name} not found !!" unless config

          ext = File.extname(name)[1..-1]
          base = name.split('.')[0]
          base + '-' + (config[:version] || config[:stamp]) + '.' + ext
        end
      end

      # retrieve and cache the full asset name
      def get_asset_version_name(name)
        cache[name] ||= begin
          config = get_config(name)
          raise "ERROR : config file for asset:#{name} not found !!" unless config

          ext = File.extname(name)[1..-1]
          base = name.split('.')[0]
          base + '-' + config[:version] + '.' + ext
        end
      end

      def set_path_root(root)
        root = root.to_s
        root = '/' + root if root[0, 1] != '/'
        root += '/' if root[-1, 1] != '/'
        @path_root = root
        @path_root_length = @path_root.length - 1
      end

      def path_root
        @path_root || '/'
      end

      attr_reader :path_root_length

      # calculate the full path of the asset
      def full_path(path)
        path = path[1..-1] if path[0, 1] == '/'
        path_root + path
      end

      # caculate the correct asset path depending on the mode of execution
      def asset_path(path, mode_production = false)
        partial_path = full_path(path)
        if mode_production
          asset_name = File.basename(partial_path)
          dir_name = File.dirname(partial_path)
          if dir_name == '.'
            AssetManager.get_asset_name(asset_name)
          else
            File.join(dir_name, AssetManager.get_asset_name(asset_name))
          end
        else
          partial_path
        end
      end

      # return the location of the compiled asset
      def file_path(path)
        name = File.basename(path)
        config = get_config(name)
        raise "ERROR : config file for asset:#{name} not found !!" unless config

        ext = File.extname(name)[1..-1]
        base = name.split('.')[0]
        config[:destination] + '/' + base + '-' + (config[:version] || config[:stamp]) + '.' + ext
      end

      # write a complete asset tag html string
      def asset_tag(partial_path, mode_production = false, opts = {})
        name = File.basename(partial_path)
        type = File.extname(partial_path)
        inline = opts.delete(:inline)
        opts_str = opts.to_a.map { |v| "#{v[0]}=\"#{v[1]}\"" }.join(' ')

        if type == '.css'
          if mode_production && inline
            content = cache['///' + name] ||= File.read(file_path(partial_path))
            %(<style type="text/css" #{opts_str} >\n#{content}</style>)
          else
            %(<link rel="stylesheet" href="#{asset_path(partial_path, mode_production)}" type="text/css" #{opts_str} >)
          end
        elsif type == '.js'
          if mode_production && inline
            # insert js here if in production mode
            content = cache['///' + name] ||= File.read(file_path(partial_path))
            %(<script>\n#{content}</script>)
          else
            %(<script src="#{asset_path(partial_path, mode_production)}" type="text/javascript"></script>)
          end
        else
          raise "unknown tag for file extension:#{partial_path}"
        end
      end

      # write an asset to the destination directory
      def write_asset(destination, name, str, opts = {})
        opts = opts.dup
        opts[:rewrite] = true unless opts.key?(:rewrite)
        raise "asset destination directory must exist:#{destination}" unless Dir.exist?(destination)

        puts "#{red('WARNING !')} Use a relative path to the destination directory" if destination[0] == '/' # with colour red
        if_modified(destination, name, str, opts) do |a|
          outfile = File.join(destination, a.newname)
          File.write outfile, str
          if a.oldname && (a.oldname != a.newname)
            oldfilename = File.join(destination, a.oldname)
            File.unlink oldfilename if File.exist?(oldfilename)
          end
          outfile
        end
      end

      private

      # escape text for printing in red on console.
      def red(text)
        "\e[31m#{text}\e[0m"
      end

    end

  end
end
